@@ -53,7 +53,6 @@ class DataController:
     def __init__(self, session, lock):
         super().__init__()
         self.session = session
-        self.evaluator_cls = Evaluator
         if session.executable:
             self.executable = Executable.deserialise(session.executable)
         self.lock = lock
@@ -148,30 +147,14 @@ class DataController:
 
     @property
     def stdout(self):
-        return [m.stdout for m in self.session.machines]
+        return list(self.session.stdout)
 
+    def write_stdout(self, value: str):
+        # don't use isinstance - it must be an actual str
+        if type(value) != str:
+            raise ValueError(f"{value} ({type(value)}) is not str")
 
-class Evaluator:
-    def __init__(self, parent):
-        self.vmid = parent.vmid
-        self.state = parent.state
-        self.data_controller = parent.data_controller
-
-    @singledispatchmethod
-    def evali(self, i: Instruction):
-        """Evaluate instruction"""
-        assert isinstance(i, Instruction)
-        raise NotImplementedError(i)
-
-    @evali.register
-    def _(self, i: mi.Print):
-        # Leave the value in the stack - print returns itself
-        val = self.state.ds_peek(0)
-        t = time.time()
-        line = f"{t:5.5f} | {val}"
-        with self.data_controller.lock:
-            self.data_controller.session.machines[self.vmid].stdout.append(line)
-
-    @evali.register
-    def _(self, i: mi.Future):
-        raise NotImplementedError
+        # Avoid empty strings
+        if value:
+            with self.lock:
+                self.session.stdout.append(value)
