@@ -11,8 +11,10 @@ import builtins
 import importlib
 import logging
 import os
+import sys
 import time
 from functools import singledispatchmethod
+from io import StringIO
 from typing import Any, Dict, List
 
 from . import types as mt
@@ -81,12 +83,14 @@ class TlMachine:
     """
 
     builtins = {
+        "wait": Wait,
         "print": Print,
         "sleep": Sleep,
         "atomp": Atomp,
         "nullp": Nullp,
         "list": List,
         "conc": Conc,
+        "append": Append,
         "first": First,
         "rest": Rest,
         # "future": Future,
@@ -97,6 +101,7 @@ class TlMachine:
         "*": Multiply,
         ">": GreaterThan,
         "<": LessThan,
+        "parse_float": ParseFloat,
     }
 
     def __init__(self, vmid, invoker):
@@ -106,7 +111,6 @@ class TlMachine:
         self.state = self.data_controller.get_state(self.vmid)
         self.probe = self.data_controller.get_probe(self.vmid)
         self.exe = self.data_controller.executable
-        self.evaluator = self.data_controller.evaluator_cls(self)
         self._foreign = {
             name: import_python_function(val.identifier, val.module)
             for name, val in self.exe.bindings.items()
@@ -153,8 +157,7 @@ class TlMachine:
     @singledispatchmethod
     def evali(self, i: Instruction):
         """Evaluate instruction"""
-        assert isinstance(i, Instruction)
-        self.evaluator.evali(i)
+        raise NotImplementedError(i)
 
     @evali.register
     def _(self, i: Bind):
@@ -183,7 +186,8 @@ class TlMachine:
         elif ptr in TlMachine.builtins:
             val = mt.TlInstruction(ptr)
         else:
-            raise Exception(f"Nothing bound to {ptr}")
+            # TODO: runtime_error(self, i, ...)
+            raise NameError(f"'{ptr}' is not defined")
 
         self.state.ds_push(val)
 
@@ -245,9 +249,18 @@ class TlMachine:
             # waiting for in the continuation
 
             py_args = list(map(mt.to_py_type, args))
-            py_result = foreign_f(*py_args)
-            result = mt.to_teal_type(py_result)
 
+            # capture Python's standard output
+            sys.stdout = capstdout = StringIO()
+            try:
+                py_result = foreign_f(*py_args)
+            finally:
+                sys.stdout = sys.__stdout__
+
+            out = capstdout.getvalue()
+            self.data_controller.write_stdout(out)
+
+            result = mt.to_teal_type(py_result)
             self.state.ds_push(result)
 
         elif isinstance(fn, mt.TlInstruction):
@@ -313,12 +326,13 @@ class TlMachine:
     @evali.register
     def _(self, i: Atomp):
         val = self.state.ds_pop()
-        self.state.ds_push(not isinstance(val, list))
+        self.state.ds_push(tl_bool(not isinstance(val, list)))
 
     @evali.register
     def _(self, i: Nullp):
         val = self.state.ds_pop()
-        self.state.ds_push(len(val) == 0)
+        isnull = isinstance(val, mt.TlNull) or len(val) == 0
+        self.state.ds_push(tl_bool(isnull))
 
     @evali.register
     def _(self, i: List):
@@ -342,16 +356,40 @@ class TlMachine:
         else:
             self.state.ds_push(mt.TlList([a] + b))
 
+    @evali.register
+    def _(self, i: Append):
+        b = self.state.ds_pop()
+        a = self.state.ds_pop()
+
+        a = mt.TlList([]) if isinstance(a, mt.TlNull) else a
+
+        if not isinstance(a, mt.TlList):
+            raise Exception(f"{a} ({type(a)}) is not a list")
+
+        self.state.ds_push(mt.TlList(a + [b]))
+
     @evali.register
     def _(self, i: First):
         lst = self.state.ds_pop()
+        if not isinstance(lst, mt.TlList):
+            raise Exception(f"{lst} ({type(lst)}) is not a list")
         self.state.ds_push(lst[0])
 
     @evali.register
     def _(self, i: Rest):
         lst = self.state.ds_pop()
+        if not isinstance(lst, mt.TlList):
+            raise Exception(f"{lst} ({type(lst)}) is not a list")
         self.state.ds_push(lst[1:])
 
+    @evali.register
+    def _(self, i: Nth):
+        n = self.state.ds_pop()
+        lst = self.state.ds_pop()
+        if not isinstance(lst, mt.TlList):
+            raise Exception(f"{lst} ({type(lst)}) is not a list")
+        self.state.ds_push(lst[n])
+
     @evali.register
     def _(self, i: Plus):
         a = self.state.ds_pop()
@@ -384,11 +422,24 @@ class TlMachine:
         b = self.state.ds_pop()
         self.state.ds_push(tl_bool(a < b))
 
+    @evali.register
+    def _(self, i: ParseFloat):
+        x = self.state.ds_pop()
+        self.state.ds_push(mt.TlFloat(float(x)))
+
     @evali.register
     def _(self, i: Sleep):
         t = self.state.ds_peek(0)
         time.sleep(t)
 
+    @evali.register
+    def _(self, i: Print):
+        # Leave the value in the stack - print() 'returns' the value printed
+        val = self.state.ds_peek(0)
+        # This should take a vmid - data stored is a tuple (vmid, str)
+        # Could also store a timestamp...
+        self.data_controller.write_stdout(str(val) + "\n")
+
     def __repr__(self):
         return f"<Machine {id(self)}>"
 
